@@ -5,7 +5,7 @@ import os.path
 from pathlib import Path
 import re
 from threading import Lock
-from time import sleep
+from time import sleep, time
 from typing import (
     Any,
     Callable,
@@ -21,15 +21,28 @@ from urllib.parse import urlparse, urlunparse
 from xml.etree.ElementTree import fromstring
 
 import click
+from dandischema import models
 from pydantic import BaseModel, Field, PrivateAttr
 import requests
 import tenacity
 
 from . import get_logger
-from .consts import MAX_CHUNK_SIZE, known_instances, known_instances_rev
-from .exceptions import NotFoundError
+from .consts import (
+    DRAFT,
+    MAX_CHUNK_SIZE,
+    DandiInstance,
+    known_instances,
+    known_instances_rev,
+)
+from .exceptions import NotFoundError, SchemaVersionError
 from .keyring import keyring_lookup
-from .utils import USER_AGENT, is_interactive, try_multiple
+from .utils import (
+    USER_AGENT,
+    check_dandi_version,
+    ensure_datetime,
+    is_interactive,
+    try_multiple,
+)
 
 lgr = get_logger()
 
@@ -44,9 +57,9 @@ class RESTFullAPIClient:
         self.api_url = api_url
         if session is None:
             session = requests.Session()
+        session.headers["User-Agent"] = USER_AGENT
         if headers is not None:
             session.headers.update(headers)
-        session.headers.setdefault("User-Agent", USER_AGENT)
         self.session = session
 
     def __enter__(self):
@@ -130,7 +143,7 @@ class RESTFullAPIClient:
             doretry |= retry
 
         try:
-            result = try_multiple(5, doretry, 1.1)(
+            result = try_multiple(12, doretry, 1.25)(
                 f,
                 url,
                 params=params,
@@ -221,6 +234,7 @@ class RESTFullAPIClient:
 
 class DandiAPIClient(RESTFullAPIClient):
     def __init__(self, api_url=None, token=None):
+        check_dandi_version()
         if api_url is None:
             instance_name = os.environ.get("DANDI_INSTANCE", "dandi")
             api_url = known_instances[instance_name].api
@@ -230,6 +244,17 @@ class DandiAPIClient(RESTFullAPIClient):
         if token is not None:
             self.authenticate(token)
 
+    @classmethod
+    def for_dandi_instance(
+        cls, instance: Union[str, DandiInstance], token=None, authenticate=False
+    ) -> "DandiAPIClient":
+        if isinstance(instance, str):
+            instance = known_instances[instance]
+        client = cls(instance.api, token=token)
+        if token is None and authenticate:
+            client.dandi_authenticate()
+        return client
+
     def authenticate(self, token):
         # Fails if token is invalid:
         self.get("/auth/token", headers={"Authorization": f"token {token}"})
@@ -269,21 +294,27 @@ class DandiAPIClient(RESTFullAPIClient):
                 break
 
     def get_dandiset(
-        self, dandiset_id: str, version_id: Optional[str] = None
+        self, dandiset_id: str, version_id: Optional[str] = None, lazy: bool = True
     ) -> "RemoteDandiset":
         """
         Fetches the Dandiset with the given ``dandiset_id``.  If ``version_id``
         is not specified, the `RemoteDandiset`'s version is set to the most
         recent published version if there is one, otherwise to the draft
         version.
+
+        If ``lazy`` is true, no requests are actually made until any data is
+        requested from the `RemoteDandiset`.
         """
-        d = RemoteDandiset._make(self, self.get(f"/dandisets/{dandiset_id}/"))
-        if version_id is not None and version_id != d.version_id:
-            if version_id == "draft":
-                d.version = d.draft_version
-            else:
-                return d.for_version(version_id)
-        return d
+        if lazy:
+            return RemoteDandiset(self, dandiset_id, version_id)
+        else:
+            d = RemoteDandiset._make(self, self.get(f"/dandisets/{dandiset_id}/"))
+            if version_id is not None and version_id != d.version_id:
+                if version_id == DRAFT:
+                    return d.for_version(d.draft_version)
+                else:
+                    return d.for_version(version_id)
+            return d
 
     def get_dandisets(self) -> Iterator["RemoteDandiset"]:
         for data in self.paginate("/dandisets/"):
@@ -295,6 +326,26 @@ class DandiAPIClient(RESTFullAPIClient):
             self, self.post("/dandisets/", json={"name": name, "metadata": metadata})
         )
 
+    def check_schema_version(self, schema_version: Optional[str] = None) -> None:
+        if schema_version is None:
+            schema_version = models.get_schema_version()
+        server_info = self.get("/info/")
+        server_schema_version = server_info.get("schema_version")
+        if not server_schema_version:
+            raise RuntimeError(
+                "Server did not provide schema_version in /info/;"
+                f" returned {server_info!r}"
+            )
+        if server_schema_version != schema_version:
+            raise SchemaVersionError(
+                f"Server requires schema version {server_schema_version};"
+                f" client only supports {schema_version}.  You may need to"
+                " upgrade dandi and/or dandischema."
+            )
+
+    def get_asset(self, asset_id: str) -> "BaseRemoteAsset":
+        return BaseRemoteAsset._from_metadata(self, self.get(f"/assets/{asset_id}"))
+
 
 class APIBase(BaseModel):
     """Base class for API objects"""
@@ -325,24 +376,87 @@ class Version(APIBase):
     modified: datetime
 
 
-class RemoteDandiset(APIBase):
+class RemoteDandiset:
     """
     Representation of a Dandiset (as of a certain version) retrieved from the
     API
     """
 
-    client: "DandiAPIClient"
-    identifier: str
-    created: datetime
-    modified: datetime
-    #: The version in question of the Dandiset
-    version: Version
-    most_recent_published_version: Optional[Version]
-    draft_version: Version
+    def __init__(
+        self,
+        client: DandiAPIClient,
+        identifier: str,
+        version: Union[str, Version, None] = None,
+        data: Optional[Dict[str, Any]] = None,
+    ) -> None:
+        self.client = client
+        self.identifier = identifier
+        self._version_id: Optional[str]
+        self._version: Optional[Version]
+        if version is None:
+            self._version_id = None
+            self._version = None
+        elif isinstance(version, str):
+            self._version_id = version
+            self._version = None
+        else:
+            self._version_id = version.identifier
+            self._version = version
+        self._data = data
+
+    def _get_data(self) -> Dict[str, Any]:
+        if self._data is None:
+            self._data = self.client.get(f"/dandisets/{self.identifier}/")
+        return self._data
 
     @property
     def version_id(self) -> str:
-        return self.version.identifier
+        if self._version_id is None:
+            self._version_id = self.version.identifier
+        return self._version_id
+
+    @property
+    def version(self) -> Version:
+        """The version in question of the Dandiset"""
+        if self._version is None:
+            if self._version_id is None:
+                self._get_data()
+            if self._data is not None:
+                for vattr in ["most_recent_published_version", "draft_version"]:
+                    vdict = self._data.get(vattr)
+                    if vdict and (
+                        self._version_id is None or vdict["version"] == self.version_id
+                    ):
+                        self._version = Version.parse_obj(vdict)
+                        self._version_id = self._version.identifier
+                        return self._version
+            assert self._version_id is not None
+            self._version = self.get_version(self._version_id)
+        return self._version
+
+    @property
+    def created(self) -> datetime:
+        return ensure_datetime(self._get_data()["created"])
+
+    @property
+    def modified(self) -> datetime:
+        return ensure_datetime(self._get_data()["modified"])
+
+    @property
+    def contact_person(self) -> str:
+        return self._get_data()["contact_person"]
+
+    @property
+    def most_recent_published_version(self) -> Optional[Version]:
+        v = self._get_data().get("most_recent_published_version")
+        if v is None:
+            return None
+        else:
+            return Version.parse_obj(v)
+
+    @property
+    def draft_version(self) -> Version:
+        return Version.parse_obj(self._get_data()["draft_version"])
 
     @property
     def api_path(self) -> str:
@@ -360,10 +474,12 @@ class RemoteDandiset(APIBase):
         as the Dandiset's version; otherwise, use ``"draft_version"``.
         """
         if data.get("most_recent_published_version") is not None:
-            version = data["most_recent_published_version"]
+            version = Version.parse_obj(data["most_recent_published_version"])
         else:
-            version = data["draft_version"]
-        return cls(client=client, version=version, **data)
+            version = Version.parse_obj(data["draft_version"])
+        return cls(
+            client=client, identifier=data["identifier"], version=version, data=data
+        )
 
     def _mkasset(self, data: Dict[str, Any]) -> "RemoteAsset":
         return RemoteAsset(
@@ -385,6 +501,33 @@ class RemoteDandiset(APIBase):
             _metadata=metadata,
         )
 
+    def json_dict(self) -> Dict[str, Any]:
+        """
+        Convert to a JSONable `dict`, omitting the ``client`` attribute and
+        using the same field names as in the API
+        """
+        data = {
+            **self._get_data(),
+            "version": self.version.json_dict(),
+            "draft_version": self.draft_version.json_dict(),
+        }
+        if self.most_recent_published_version is not None:
+            data[
+                "most_recent_published_version"
+            ] = self.most_recent_published_version.json_dict()
+        return data
+
+    def refresh(self) -> None:
+        """
+        Update the `RemoteDandiset` in-place with the latest data from the
+        server.  The `RemoteDandiset` continues to have the same version as
+        before, but the cached version data is internally cleared and may be
+        different upon subsequent access.
+        """
+        self._data = self.client.get(f"/dandisets/{self.identifier}/")
+        # Clear _version so it will be refetched the next time it is accessed
+        self._version = None
+
     def get_versions(self) -> Iterator[Version]:
         """Returns an iterator of all available `Version`\\s for the Dandiset"""
         for v in self.client.paginate(f"{self.api_path}versions/"):
@@ -396,7 +539,9 @@ class RemoteDandiset(APIBase):
         version does not exist, a `requests.HTTPError` is raised with a 404
         status code.
         """
-        return Version.parse_obj(self.client.get(f"{self.version_api_path}info/"))
+        return Version.parse_obj(
+            self.client.get(f"/dandisets/{self.identifier}/versions/{version_id}/info")
+        )
 
     def for_version(self, version_id: Union[str, Version]) -> "RemoteDandiset":
         """
@@ -407,11 +552,28 @@ class RemoteDandiset(APIBase):
         """
         if isinstance(version_id, str):
             version_id = self.get_version(version_id)
-        return self.copy(update={"version": version_id})
+        return type(self)(
+            client=self.client,
+            identifier=self.identifier,
+            version=version_id,
+            data=self._data,
+        )
 
     def delete(self) -> None:
-        """Delete the Dandiset"""
+        """
+        Delete the Dandiset from the server.  Any further access of the
+        instance's data attributes afterwards will result in a 404.
+        """
         self.client.delete(self.api_path)
+        self._data = None
+        self._version = None
+
+    def get_metadata(self) -> models.Dandiset:
+        """
+        Fetch the metadata for this version of the Dandiset as a
+        `dandischema.models.Dandiset` instance
+        """
+        return models.Dandiset.parse_obj(self.get_raw_metadata())
 
     def get_raw_metadata(self) -> Dict[str, Any]:
         """
@@ -420,6 +582,12 @@ class RemoteDandiset(APIBase):
         """
         return cast(Dict[str, Any], self.client.get(self.version_api_path))
 
+    def set_metadata(self, metadata: models.Dandiset) -> None:
+        """
+        Set the metadata for this version of the Dandiset to the given value
+        """
+        self.set_raw_metadata(metadata.json_dict())
+
     def set_raw_metadata(self, metadata: Dict[str, Any]) -> None:
         """
         Set the metadata for this version of the Dandiset to the given value
@@ -429,35 +597,37 @@ class RemoteDandiset(APIBase):
             json={"metadata": metadata, "name": metadata.get("name", "")},
         )
 
+    def wait_until_valid(self, min_time=20):
+        """
+        Wait for a Dandiset to be valid.  Validation is a background celery
+        task which runs asynchronously, so we need to wait for it to complete.
+        """
+        lgr.debug("Waiting for Dandiset %s to complete validation ...", self.identifier)
+        start = time()
+        while time() - start < min_time:
+            r = self.client.get(f"{self.version_api_path}info/")
+            if "status" not in r:
+                # Running against older version of dandi-api that doesn't
+                # validate
+                return
+            if r["status"] == "Valid":
+                return
+            sleep(0.5)
+        raise ValueError(
+            f"Dandiset {self.identifier} is {r['status']}: {r['validation_error']}"
+        )
+
     def publish(self) -> "RemoteDandiset":
         """
         Publish this version of the Dandiset.  Returns a copy of the
         `RemoteDandiset` with the `version` attribute set to the new published
         `Version`.
         """
-        lgr.debug("Waiting for Dandiset %s to complete validation ...", self.identifier)
-        while True:
-            r = self.client.get(f"{self.version_api_path}info/")
-            if "status" not in r:
-                # Running against older version of dandi-api that doesn't
-                # validate
-                break
-            if r["status"] == "Valid":
-                break
-            elif r["status"] == "Invalid":
-                raise ValueError(
-                    f"Dandiset {self.identifier} is invalid: {r['validation_error']}"
-                )
-            sleep(0.5)
-        return self.copy(
-            update={
-                "version": Version.parse_obj(
-                    self.client.post(f"{self.version_api_path}publish/")
-                )
-            }
+        return self.for_version(
+            Version.parse_obj(self.client.post(f"{self.version_api_path}publish/"))
         )
 
-    def get_assets(self, path=None) -> Iterator["RemoteAsset"]:
+    def get_assets(self) -> Iterator["RemoteAsset"]:
         """Returns an iterator of all assets in this version of the Dandiset"""
         for a in self.client.paginate(f"{self.version_api_path}assets/"):
             yield self._mkasset(a)
@@ -472,7 +642,7 @@ class RemoteDandiset(APIBase):
             self.client.get(f"{self.version_api_path}assets/{asset_id}/")
         )
 
-    def get_assets_under_path(self, path: str) -> Iterator["RemoteAsset"]:
+    def get_assets_with_path_prefix(self, path: str) -> Iterator["RemoteAsset"]:
         """
         Returns an iterator of all assets in this version of the Dandiset whose
         `~RemoteAsset.path` attributes start with ``path``
@@ -491,7 +661,9 @@ class RemoteDandiset(APIBase):
         try:
             # Weed out any assets that happen to have the given path as a
             # proper prefix:
-            (asset,) = (a for a in self.get_assets_under_path(path) if a.path == path)
+            (asset,) = (
+                a for a in self.get_assets_with_path_prefix(path) if a.path == path
+            )
         except ValueError:
             raise NotFoundError(f"No asset at path {path!r}")
         else:
@@ -509,7 +681,7 @@ class RemoteDandiset(APIBase):
         """
         if assets_dirpath and not assets_dirpath.endswith("/"):
             assets_dirpath += "/"
-        assets = list(self.get_assets_under_path(assets_dirpath))
+        assets = list(self.get_assets_with_path_prefix(assets_dirpath))
         for a in assets:
             filepath = Path(dirpath, a.path[len(assets_dirpath) :])
             filepath.parent.mkdir(parents=True, exist_ok=True)
@@ -520,24 +692,26 @@ class RemoteDandiset(APIBase):
         filepath: Union[str, Path],
         asset_metadata: Dict[str, Any],
         jobs: Optional[int] = None,
+        replace_asset: Optional["RemoteAsset"] = None,
     ) -> "RemoteAsset":
         """
         Upload the file at ``filepath`` with metadata ``asset_metadata`` to
         this version of the Dandiset and return the resulting asset.  Blocks
         until the upload is complete.
 
-        Parameters
-        ----------
-        filepath: str or PathLike
-          the path to the local file to upload
-        asset_metadata: dict
-          Metadata for the uploaded asset file.  Must include a "path" field
-          giving the POSIX path at which the uploaded file will be placed on
-          the server.
-        jobs: int
-          Number of threads to use for uploading; defaults to 5
+        :param filepath: the path to the local file to upload
+        :type filepath: str or PathLike
+        :param dict asset_metadata:
+            Metadata for the uploaded asset file.  Must include a "path" field
+            giving the POSIX path at which the uploaded file will be placed on
+            the server.
+        :param int jobs: Number of threads to use for uploading; defaults to 5
+        :param RemoteAsset replace_asset: If set, replace the given asset,
+            which must have the same path as the new asset
         """
-        for status in self.iter_upload_raw_asset(filepath, asset_metadata, jobs=jobs):
+        for status in self.iter_upload_raw_asset(
+            filepath, asset_metadata, jobs=jobs, replace_asset=replace_asset
+        ):
             if status["status"] == "done":
                 return status["asset"]
         raise RuntimeError("iter_upload_raw_asset() finished without returning 'done'")
@@ -547,28 +721,28 @@ class RemoteDandiset(APIBase):
         filepath: Union[str, Path],
         asset_metadata: Dict[str, Any],
         jobs: Optional[int] = None,
+        replace_asset: Optional["RemoteAsset"] = None,
     ) -> Iterator[dict]:
         """
         Upload the file at ``filepath`` with metadata ``asset_metadata`` to
         this version of the Dandiset, returning a generator of status
         `dict`\\s.
 
-        Parameters
-        ----------
-        filepath: str or PathLike
-          the path to the local file to upload
-        asset_metadata: dict
-          Metadata for the uploaded asset file.  Must include a "path" field
-          giving the POSIX path at which the uploaded file will be placed on
-          the server.
-        jobs: int
-          Number of threads to use for uploading; defaults to 5
-
-        Returns
-        -------
-        A generator of `dict`\\s containing at least a ``"status"`` key.  Upon
-        successful upload, the last `dict` will have a status of ``"done"`` and
-        an ``"asset"`` key containing the resulting `RemoteAsset`.
+        :param filepath: the path to the local file to upload
+        :type filepath: str or PathLike
+        :param dict asset_metadata:
+            Metadata for the uploaded asset file.  Must include a "path" field
+            giving the POSIX path at which the uploaded file will be placed on
+            the server.
+        :param int jobs:
+            Number of threads to use for uploading; defaults to 5
+        :param RemoteAsset replace_asset: If set, replace the given asset,
+            which must have the same path as the new asset
+        :returns:
+            A generator of `dict`\\s containing at least a ``"status"`` key.
+            Upon successful upload, the last `dict` will have a status of
+            ``"done"`` and an ``"asset"`` key containing the resulting
+            `RemoteAsset`.
         """
         from .support.digests import get_dandietag
 
@@ -674,20 +848,18 @@ class RemoteDandiset(APIBase):
                 blob_id = resp["blob_id"]
         lgr.debug("%s: Assigning asset blob to dandiset & version", asset_path)
         yield {"status": "producing asset"}
-        try:
-            extant = self.get_asset_by_path(asset_path)
-        except NotFoundError:
+        if replace_asset is not None:
+            lgr.debug("%s: Replacing pre-existing asset")
             a = self._mkasset(
-                self.client.post(
-                    f"{self.version_api_path}assets/",
+                self.client.put(
+                    replace_asset.api_path,
                     json={"metadata": asset_metadata, "blob_id": blob_id},
                 )
             )
         else:
-            lgr.debug("%s: Asset already exists at path; updating", asset_path)
             a = self._mkasset(
-                self.client.put(
-                    extant.api_path,
+                self.client.post(
+                    f"{self.version_api_path}assets/",
                     json={"metadata": asset_metadata, "blob_id": blob_id},
                 )
             )
@@ -695,17 +867,9 @@ class RemoteDandiset(APIBase):
         yield {"status": "done", "asset": a}
 
 
-class RemoteAsset(APIBase):
-    """Representation of an asset retrieved from the API"""
-
-    JSON_EXCLUDE = frozenset(["client", "dandiset_id", "version_id"])
-
+class BaseRemoteAsset(APIBase):
     client: "DandiAPIClient"
-    #: The identifier for the Dandiset to which the asset belongs
-    dandiset_id: str
-    #: The identifier for the version of the Dandiset to which the asset
-    #: belongs
-    version_id: str
+
     #: The asset identifier
     identifier: str = Field(alias="asset_id")
     path: str
@@ -721,14 +885,34 @@ class RemoteAsset(APIBase):
         # underscores, so we have to do it ourselves.
         self._metadata = data.get("metadata", data.get("_metadata"))
 
+    @classmethod
+    def _from_metadata(
+        self, client: "DandiAPIClient", metadata: Dict[str, Any]
+    ) -> "BaseRemoteAsset":
+        return BaseRemoteAsset(
+            client=client,
+            identifier=metadata["identifier"],
+            path=metadata["path"],
+            size=metadata["contentSize"],
+            modified=metadata["dateModified"],
+            _metadata=metadata,
+        )
+
     @property
     def api_path(self) -> str:
-        return f"/dandisets/{self.dandiset_id}/versions/{self.version_id}/assets/{self.identifier}/"
+        return f"/assets/{self.identifier}/"
 
     @property
     def download_url(self) -> str:
         return self.client.get_url(f"{self.api_path}download/")
 
+    def get_metadata(self) -> models.Asset:
+        """
+        Fetch the metadata for the asset as a `dandischema.models.Asset`
+        instance
+        """
+        return models.Asset.parse_obj(self.get_raw_metadata())
+
     def get_raw_metadata(self) -> Dict[str, Any]:
         """Fetch the metadata for the asset as an unprocessed `dict`"""
         if self._metadata is not None:
@@ -736,6 +920,22 @@ class RemoteAsset(APIBase):
         else:
             return cast(Dict[str, Any], self.client.get(self.api_path))
 
+    def get_digest(
+        self, digest_type: Union[str, models.DigestType] = models.DigestType.dandi_etag
+    ) -> str:
+        """
+        Retrieves the value of the given type of digest from the asset's
+        metadata.  Raises `NotFoundError` if there is no entry for the given
+        digest type.
+        """
+        if isinstance(digest_type, models.DigestType):
+            digest_type = digest_type.value
+        metadata = self.get_raw_metadata()
+        try:
+            return metadata["digest"][digest_type]
+        except KeyError:
+            raise NotFoundError(f"No {digest_type} digest found in metadata")
+
     def get_content_url(
         self,
         regex: str = r".*",
@@ -780,10 +980,6 @@ class RemoteAsset(APIBase):
             url = urlunparse(urlparse(url)._replace(query=""))
         return url
 
-    def delete(self) -> None:
-        """Delete the asset"""
-        self.client.delete(self.api_path)
-
     def get_download_file_iter(
         self, chunk_size: int = MAX_CHUNK_SIZE
     ) -> Callable[..., Iterator[bytes]]:
@@ -823,6 +1019,58 @@ class RemoteAsset(APIBase):
                 fp.write(chunk)
 
 
+class RemoteAsset(BaseRemoteAsset):
+    """
+    Representation of an asset retrieved from the API with associated Dandiset
+    information
+    """
+
+    JSON_EXCLUDE = frozenset(["client", "dandiset_id", "version_id"])
+
+    #: The identifier for the Dandiset to which the asset belongs
+    dandiset_id: str
+    #: The identifier for the version of the Dandiset to which the asset
+    #: belongs
+    version_id: str
+
+    @property
+    def api_path(self) -> str:
+        return f"/dandisets/{self.dandiset_id}/versions/{self.version_id}/assets/{self.identifier}/"
+
+    def set_metadata(self, metadata: models.Asset) -> None:
+        """
+        Set the metadata for the asset to the given value and update the
+        `RemoteAsset` in place.
+        """
+        return self.set_raw_metadata(metadata.json_dict())
+
+    def set_raw_metadata(self, metadata: Dict[str, Any]) -> None:
+        """
+        Set the metadata for the asset to the given value and update the
+        `RemoteAsset` in place.
+        """
+        try:
+            etag = metadata["digest"]["dandi:dandi-etag"]
+        except KeyError:
+            raise ValueError("dandi-etag digest not set in new asset metadata")
+        r = self.client.post(
+            "/blobs/digest/",
+            json={"algorithm": "dandi:dandi-etag", "value": etag},
+        )
+        data = self.client.put(
+            self.api_path, json={"metadata": metadata, "blob_id": r["blob_id"]}
+        )
+        self.identifier = data["asset_id"]
+        self.path = data["path"]
+        self.size = int(data["size"])
+        self.modified = ensure_datetime(data["modified"])
+        self._metadata = data["metadata"]
+
+    def delete(self) -> None:
+        """Delete the asset"""
+        self.client.delete(self.api_path)
+
+
 def upload_part(storage_session, fp, lock, etagger, asset_path, part):
     etag_part = etagger.get_part(part["part_number"])
     if part["size"] != etag_part.size:
